# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A rag-bag of utility functions that did not fit anywhere else."""

from collections import defaultdict
import contextlib
import functools
import io
import os
import pkgutil
import re
import shutil
import sys

RE_BRACKETS = re.compile('{[^{]*}')

def copy(source, destination):
    """Copy ``source`` to ``destination``, creating directories if necessary."""
    os.makedirs(os.path.dirname(destination), exist_ok=True)
    shutil.copy2(source, destination)

def copytree(source, destination):
    """Recursively copy a tree.

    Note: We do not use :func:`shutil.copytree` because it raises an error if
    destination already exists, which we do not want.
    """
    os.makedirs(source, exist_ok=True)
    for path in os.listdir(source):
        sourceitem = os.path.join(source, path)
        destitem = os.path.join(destination, path)
        if os.path.isdir(sourceitem):
            copytree(sourceitem, destitem)
        else:
            copy(sourceitem, destitem)

def partial_format(string, *args, **kwargs):
    """Format string, leaving unfilled fields untouched.

    >>> partial_format("To {verb} or {negation} to {verb}", negation="not")
    'To {verb} or not to {verb}'
    """
    dictionary = {
        field[1:-1]: field
        for field
        in re.findall(RE_BRACKETS, string)
        }

    dictionary.update(kwargs)
    return string.format(*args, **dictionary)

@contextlib.contextmanager
def smart_open(filename, mode="w"):
    """Open filename, standard output, or nothing.

    If ``filename`` is:

    * ``""`` (the empty string): return a fake file object, which is empty if
      file is open for reading, can be written in if file is open for writing;
    * ``"-"`` (a dash): read from standard input, or write to standard output
      (depending on mode);
    * any other: open the given file.
    """
    if (not filename) or (not filename.strip()):
        with io.StringIO() as file:
            yield file
    elif filename.strip() == "-":
        if 'w' in mode:
            yield sys.stdout
        elif 'r' in mode:
            yield sys.stdin
    else:
        with open(filename, mode) as file:
            yield file


def yesno(arg):
    """Interpret some variable as a boolean.

    >>> yesno("y")
    True
    >>> yesno("0")
    False
    >>> yesno("1")
    True
    >>> yesno("Yes")
    True
    >>> yesno("True")
    True
    >>> yesno("something senseless")
    False
    >>> yesno(None)
    False
    """
    if isinstance(arg, bool):
        return arg
    try:
        return bool(int(arg))
    except (ValueError, TypeError):
        pass
    if isinstance(arg, str):
        return (arg.lower() in ['y', 'yes', 'true'])
    return bool(arg)

class ChangeDir:
    """Context manager to change and restore current directory."""
    # pylint: disable=too-few-public-methods

    def __init__(self, directory):
        if not directory:
            directory = "."
        self._newdir = directory
        self._olddir = os.getcwd()

    def __enter__(self):
        self._olddir = os.getcwd()
        os.chdir(self._newdir)

    def __exit__(self, *args, **kwargs):
        os.chdir(self._olddir)

def expand_path(path):
    """Return ``path`` where environment variables and user directory have been expanded."""
    return os.path.expanduser(os.path.expandvars(path))

class DeepDict(defaultdict):
    """Dictionary of dictionary of ... of dictionaries."""

    def __init__(self, depth, dictionary=None):
        if dictionary is None:
            dictionary = {}

        self.depth = depth
        if depth == 1:
            factory = lambda: None
        else:
            factory = functools.partial(self.__class__, depth-1)

        super().__init__(factory)
        for key, value in dictionary.items():
            self[key].update(value)

    @classmethod
    def from_configparser(cls, config):
        """Create a :class:`DeepDict` object from a :class:`configparser.ConfigParser` object."""
        dictionary = cls(2)
        for section in config.sections():
            for option in config.options(section):
                dictionary[section][option] = config.get(section, option)
        return dictionary

    @property
    def __dict__(self):
        """Convert self to regular :type:`dict`."""
        if self.depth == 1:
            return dict(self)
        else:
            return {
                key: vars(value)
                for (key, value)
                in self.items()
                }

def iter_modules(path, prefix, logger=None):
    """Iterate over modules.

    :arg list path: List of paths to look for modules in.
    :arg str prefix: String to output on the front of every module name on
        output.
    :arg function logger: Function used to print logging error. May be `None`
        to discard logging.
    """
    for module_finder, name, __is_pkg in pkgutil.walk_packages(path, prefix):
        if name in sys.modules:
            module = sys.modules[name]
        else:
            try:
                module = module_finder.find_spec(name).loader.load_module()
            except ImportError as error:
                if logger is not None:
                    logger("[plugins] Could not load module {}: {}".format(name, str(error)))
                continue
        yield module



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Useless action. Nothing has changed since last compilation."""

import pathlib

from evariste.action import Action, PathCompiler, Report

class UselessCompiler(PathCompiler):
    """Does nothing to compile path (most likely because it is cached)."""
    # pylint: disable=too-few-public-methods

    def compile(self):
        cached = self.shared.tree[self.path.from_source]['misc.changed']
        if cached is None:
            target = self.path.from_source
            depends = set()
        else:
            fs_target = pathlib.Path(list(cached['target'].keys())[0])
            target = fs_target.relative_to(self.path.root.from_fs)
            depends = set([
                pathlib.Path(path)
                for path
                in cached['depends'].keys()
                ]) - set([self.path.from_fs])
        return Report(
            self.path,
            target=target,
            depends=depends,
            success=True,
            )

class Useless(Action):
    """Useless action. Nothing has changed since last compilation."""

    keyword = "useless"
    pathcompiler = UselessCompiler
    required = True



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provide hook mechanism to builder"""

class Hook:
    """Hook, containing functions to be called before and after a certain event.
    """
    # pylint: disable=too-few-public-methods

    data = None

    def __init__(self, plugin, shared):
        self.shared = shared
        self.plugin = plugin
        self.local = self.plugin.local

class ContextHook(Hook):
    """Hook executed before and after a context"""

    def pre(self, *args, **kwargs):
        # pylint: disable=no-self-use, unused-argument
        """Hook called before the context.

        Takes as arguments the arguments ``args`` and ``kwargs`` passed to the
        the :meth:`evariste.plugins.AllPlugins.contexthook` method.

        :return: Nothing.
        """
        return

    def post(self):
        # pylint: disable=no-self-use
        """Hook called after the context."""
        return

class MethodHook(Hook):
    """Hook executed before and after a method call.

    .. note::

        For both :meth:`pre` and :meth:`post` methods, do note that the
        calling order of all the functions hooked to a particular event is
        undefined. Do not assume anything about this.
    """

    def pre(self, *args, **kwargs):
        """Hook called before the method.

        Get as parameters the parameters passed to the method (including
        `self`). It can return:
        - a tuple ``args``, ``kwargs``, which will be passed to the method
          (this can be used to change the value of the arguments);
        - ``None``, which is a shortcut for "I did not change the arguments".
        """
        # pylint: disable=no-self-use
        return args, kwargs

    def post(self, value):
        """Hook called after the method."""
        # pylint: disable=no-self-use
        return value

def methodhooks(name=None, *, pre=True, post=True, getter=None):
    """Decorator to add hooks to a class method.

    :param str name: Name of the hook to register.
    :param bool pre: Enable pre-hook.
    :param bool post: Enable post-hook.
    :param function getter: Function that, given the instance object as argument,
        returns a :class:`plugins.AllPlugins` object. If ``None``, the default
        ``self.shared.builder.plugins`` is used (``self`` is supposed to have
        this attribute).
    """

    def decorator(function):
        """Actual decorator."""
        if name == None:
            hook = function.__name__
        else:
            hook = name

        def wrap(*args, **kwargs):
            """Wrapped function."""
            self = args[0]
            if getter is None:
                plugins = self.shared.builder.plugins
            else:
                plugins = getter(*args, **kwargs)
            return plugins.apply_hook(hook, function, pre=pre, post=post, args=args, kwargs=kwargs)

        return wrap

    return decorator



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Command line client to :mod:`evariste`"""

import logging
import os
import sys

from evariste import errors
from evariste.builder import Builder
from evariste.main.options import get_options
from evariste.setup import Setup
from evariste.utils import ChangeDir
import evariste

FORMAT = '[%(threadName)10s-%(name)-20s] %(message)s'
logging.basicConfig(format=FORMAT)
LOGGER = logging.getLogger(evariste.__name__)

DEFAULT_SETUP = {
    'misc.changed': {
        'time': 'fs',
        }
    }

def main():
    """Main function"""
    arguments = get_options()
    try:
        # pylint: disable=no-member
        with ChangeDir(os.path.dirname(arguments.setup)):
            default = DEFAULT_SETUP
            default.update({'arguments': vars(arguments)})
            with Builder.from_setupname(
                os.path.basename(arguments.setup),
                default=default,
            ) as builder:
                builder.compile()
                for renderer in builder.iter_renderers():
                    renderer()
    except errors.EvaristeError as error:
        LOGGER.error("Error: " + str(error))
        sys.exit(1)
    except errors.EvaristeBug as bug:
        LOGGER.error("Error: " + str(bug))
        LOGGER.error(
            "You should not see this: you just discovered a bug. Please copy "
            "the full error message and report a bug."
            )
        sys.exit(2)
    except KeyboardInterrupt:
        LOGGER.error("Aborted.")
        sys.exit(1)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Actions performed to compile files."""

import logging
import os

from evariste import errors, plugins

LOGGER = logging.getLogger(__name__)

################################################################################
# Actions

class PathCompiler:
    """Compiler of one path.

    The main difference between this object and :class:`Action` is that:
        - there is one instance of :class:`Action` for the whole tree;
        - there is one instance of :class:`PathCompiler` per path.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, parent, path, plugin_id, shared):
        self.shared = shared
        self.plugin_id = plugin_id
        self.path = path
        self.parent = parent

    def compile(self):
        """Perform actual compilation of `self.path`."""
        raise NotImplementedError()

class Action(plugins.PluginBase):
    """Generic action"""
    # pylint: disable=too-many-instance-attributes, too-few-public-methods

    plugin_type = "action"
    pathcompiler = PathCompiler

    def compile(self, path):
        """Compile ``path``, catching :class:`EvaristeError` exceptions.

        This function *must* be thread-safe.
        """
        try:
            return self.pathcompiler(self, path, self.plugin_id, self.shared).compile()
        except errors.EvaristeError as error:
            LOGGER.error(str(error))
            return Report(
                path,
                success=False,
                error=str(error),
                log=[LogInternal(str(error))],
                )

class DirectoryCompiler(PathCompiler):
    """Compiler of a directory."""

    def compile(self):
        return Report(
            self.path,
            success=self.success(),
            error="At least one file in this directory failed.",
            target=None,
            )

    def success(self):
        """Return ``True`` if compilation of all subpath succeeded."""
        for sub in self.path:
            if not self.path[sub].report.success:
                return False
        return True


class DirectoryAction(Action):
    """Fake action on directories."""
    # pylint: disable=abstract-method, too-few-public-methods

    keyword = "directory"
    pathcompiler = DirectoryCompiler
    required = True

    def match(self, dummy):
        return False

################################################################################
# Reports

class Report:
    """Report of an action. Mainly a namespace with very few methods."""

    def __init__(
            self,
            path,
            target=None,
            error="Undocumented error",
            success=False,
            log=None,
            depends=None,
        ):
        # pylint: disable=too-many-arguments

        self.depends = depends
        if self.depends is None:
            self.depends = set()

        self.log = log
        if self.log is None:
            self.log = []

        self.path = path
        self.target = target
        self._success = success
        self.error = error

    @property
    def full_depends(self):
        """Set of files this action depends on, including ``self.path``."""
        return self.depends | set([self.path.from_fs])

    @property
    def success(self):
        """Success getter"""
        return self._success

    @success.setter
    def success(self, value):
        """Success setter."""
        self._success = value




################################################################################
# Log

class Log:
    """Action log"""
    # pylint: disable=too-few-public-methods

    def __init__(self, identifier, content):
        self.identifier = identifier
        self.content = content

    def __str__(self):
        return "({}) {}".format(self.identifier, self.content)

class LogInternal(Log):
    """Log of internal exception"""
    # pylint: disable=too-few-public-methods

    def __init__(self, content):
        super().__init__("Internal error", content)

class LogFile(Log):
    """Log from file"""
    # pylint: disable=too-few-public-methods

    def __init__(self, filename):
        with open(filename, errors="replace") as file:
            super().__init__(filename, file.read())

class LogPipe(Log):
    """Log from pipe"""
    # pylint: disable=too-few-public-methods

    IDENTIFIERS = {
        0: "standard input", # Why would one do that?
        1: "standard output",
        2: "standard error",
        }

    def __init__(self, pipe, content=""):
        super().__init__(self.IDENTIFIERS.get(pipe, str(pipe)), content)

    def read(self, pipe):
        """Read content from pipe."""
        with os.fdopen(pipe, errors="replace") as file:
            for line in file:
                self.content += line

class MultiLog(Log):
    """List of logs"""
    # pylint: disable=too-few-public-methods

    def __init__(self, sublogs):
        super().__init__("multilog", "")
        self._sublogs = sublogs

    def __iter__(self):
        yield from self._sublogs




# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin to evaluate and store modification times of files.

It is used not to re-compile files that have not changed.
"""

from datetime import datetime
import os
import pathlib

from evariste.misc import Misc
from evariste.hooks import ContextHook, MethodHook
from evariste.tree import File

class MakeArchive(MethodHook):
    """Hook called before and after archive creation."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tree = None

    def pre(self, tree, *__args, **__kwargs):
        self.tree = tree

    def post(self, value):
        if self.plugin.compile(self.tree):
            self.local.tree[self.tree.from_source]['archivepath'] = value
        return value

class BuildTreeItem(MethodHook):
    """Hook called before and after the tree is built."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tree = None

    def pre(self, tree, *__args, **__kwargs):
        self.tree = tree

    def post(self, value):
        """Compare data with cache to see if a compilation is needed."""
        if not isinstance(self.tree, File):
            return value

        tree = self.tree

        # If the --always-compile argument was given to the command line call,
        # do compile.
        if self.shared.setup['arguments']['always_compile']:
            self.plugin.files_to_compile.add(tree.from_fs.resolve())
            return value

        # If tree is not in hash, compile it
        if (self.local.tree[tree.from_source] is None) or (self.local.tree[tree.from_source] == {}):
            self.plugin.files_to_compile.add(tree.from_fs.resolve())
            return value

        # Otherwise, check if data hash has changed
        for key in ["depends", "target"]:
            dependencies = self.local.tree[tree.from_source][key]
            for depend in dependencies:
                cache_mtime = dependencies[depend][self.local.setup['time']]
                if self.local.setup['time'] == 'vcs':
                    file_mtime = tree.vcs.last_modified(pathlib.Path(depend))
                else:
                    try:
                        file_mtime = datetime.fromtimestamp(os.path.getmtime(depend))
                    except FileNotFoundError:
                        self.plugin.files_to_compile.add(tree.from_fs.resolve())
                        return value
                if cache_mtime != file_mtime:
                    self.plugin.files_to_compile.add(tree.from_fs.resolve())
                    return value

        return value

class CompileFile(ContextHook):
    """Hook called before and after compilation of one file."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tree = None

    def pre(self, tree, *_args, **_kwargs):
        self.tree = tree

    def post(self):
        """Store the modification time of file and its dependencies."""
        if not isinstance(self.tree, File):
            return

        tree = self.tree
        if not tree.report.success:
            self.plugin.files_to_compile.discard(tree.from_fs.resolve())
            return

        if self.local.tree[tree.from_source] is None:
            self.local.tree[tree.from_source] = dict()

        # Saving file and dependencies data into cache
        self.local.tree[tree.from_source]['depends'] = {
            file.as_posix(): {
                'vcs': tree.vcs.last_modified(file),
                'fs': datetime.fromtimestamp(os.path.getmtime(file.as_posix()))
            }
            for file
            in tree.full_depends()
            }
        self.local.tree[tree.from_source]['target'] = {
            file.as_posix(): {
                'vcs': tree.vcs.last_modified(file),
                'fs': datetime.fromtimestamp(os.path.getmtime(file.as_posix()))
            }
            for file
            in [tree.root.from_fs / tree.report.target]
            }

        # Mark dependencies as ignored (not to be compiled)
        self.plugin.files_to_ignore.update([
            depend.resolve()
            for depend
            in tree.depends()
            ])

class Changed(Misc):
    """Plugin to evaluate and store modification times of files."""

    keyword = "changed"
    default_setup = {
        'time': 'vcs',
        }
    hooks = {
        'buildtreeitem': BuildTreeItem,
        'compilefile': CompileFile,
        'makearchive': MakeArchive,
        }
    required = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.files_to_compile = set()
        self.files_to_ignore = set()

    def compile(self, tree):
        """Return `True` iff file `tree` is to be compiled.

        It is not compiled if:
        - nothing has changed since last compilation;
        - it is a dependency of a file described in previous item.
        """
        return (
            (tree.from_fs.resolve() in self.files_to_compile)
            and
            (tree.from_fs.resolve() not in self.files_to_ignore)
            )



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract class for jinja2 renderers."""

import jinja2
import os
import pkg_resources
import pathlib
import textwrap

from evariste import utils
from evariste.renderer.dest import DestRenderer
from evariste.renderer.jinja2.readme import Jinja2ReadmeRenderer
from evariste.renderer.jinja2.target import Jinja2TargetRenderer

class JinjaRenderer(DestRenderer):
    """Abstract class for jinja2 renderers."""
    # pylint: disable=too-few-public-methods

    extension = None
    subplugins = {
        "target": Jinja2TargetRenderer,
        "readme": Jinja2ReadmeRenderer,
        }

    def __init__(self, shared):
        super().__init__(shared)

        self.environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                self._templatedirs(),
            )
        )
        self.environment.filters['basename'] = os.path.basename
        self.environment.filters['yesno'] = utils.yesno

    def _templatedirs(self):
        """Iterator over the directories in which templates may exist.

        - Directories are returned as strings;
        - directories may not exist.
        """
        if self.local.setup['templatedirs'] is not None:
            yield from utils.expand_path(self.local.setup['templatedirs']).split()
        yield pkg_resources.resource_filename( #pylint: disable=no-member
            self.__class__.__module__,
            os.path.join("data", "templates")
            )
        yield from [
            os.path.join(utils.expand_path(path), "templates")
            for path
            in [".evariste", "~/.config/evariste", "~/.evariste", "/usr/share/evariste"]
            ]

    def _template(self):
        """Return the name of the template to use."""
        if self.local.setup['template'] is not None:
            return self.local.setup['template']
        return "tree.{}".format(self.extension)

    def render(self, tree):
        return self._render(
            tree=tree,
            template=self._template(),
            context={
                'destdir': pathlib.Path(self.destdir),
                'shared': self.shared,
                'local': self.local,
                'render': self._render,
                'render_download': self._render_download,
                'render_readme': self._render_readme,
                'render_file': self._render_file,
                'render_directory': self._render_directory,
                'render_template': self._render_template,
                },
            )

    @jinja2.contextfunction
    def _render(self, context, tree, template):
        """Render the tree, using given template (or template list)."""
        if isinstance(context, dict):
            context['tree'] = tree
        else:
            context.vars['tree'] = tree
        if tree.is_file() and tree.report.success:
            utils.copy(
                (tree.root.from_fs / tree.report.target).as_posix(),
                (context['destdir'] / tree.report.target).as_posix(),
                )
        return textwrap.indent(
            self.environment.get_or_select_template(template).render(context),
            "  ",
            )

    @jinja2.contextfunction
    def _render_directory(self, context, tree):
        """Render ``tree``, which is a :class:`tree.Directory`, as HTML."""
        return self._render(context, tree, "tree_directory.{}".format(self.extension))

    @jinja2.contextfunction
    def _render_file(self, context, tree):
        """Render ``tree``, which is a :class:`tree.File`."""
        return self.shared.builder.plugins.match_plugin(
            "{}.target".format(self.plugin_id), tree
            ).render(tree, context)

    @jinja2.contextfunction
    def _render_download(self, context, tree):
        """Render the code downloading the archive."""
        path = tree.make_archive(context["destdir"])
        context.vars['href'] = path.as_posix()
        context.vars['content'] = path.name
        return self.environment.get_template("download.{}".format(self.extension)).render(context) # pylint: disable=no-member

    @jinja2.contextfunction
    def _render_readme(self, context, tree):
        """Find the readme of tree, and returns the corresponding code."""
        # pylint: disable=unused-argument
        readme_plugins = self.shared.builder.plugins.get_plugin_type(
            "{}.readme".format(self.plugin_id)
            ).values()
        readme = tree.render_readme(readme_plugins, context)
        if readme:
            return readme
        return ""

    @jinja2.contextfunction
    def _render_template(self, context, template):
        """Render template given in argument."""
        with open(template) as file:
            return self.environment.from_string(file.read()).render(context)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Text rendere: print report on terminal."""

import unidecode
import sys

from evariste.renderer import Renderer
from evariste.utils import yesno

class TextRenderer(Renderer):
    """Text renderer: print report on terminal."""

    keyword = 'text'
    default_setup = {
        'color': "auto",
        'ascii': False,
        }

    def render(self, tree):

        # Option 'ascii'
        if yesno(self.local.setup['ascii']):
            def asciiprint(text):
                """Convert argument to ascii, and print it."""
                print(unidecode.unidecode(text))
            print_function = asciiprint
        else:
            print_function = print

        # Option 'color'
        color = self.local.setup['color']
        if color.lower() == 'auto':
            color = sys.stdout.isatty() # pylint: disable=no-member
        color = yesno(color)

        # Render
        tree.pprint(
            report=True,
            color=color,
            print_function=print_function,
            )



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Cache

Implements a caching mechanism, to be able to store data between successive
runs of Évariste.
"""

import os
import pickle

from evariste.shared import Shared, SharedDeepDict
from evariste.utils import expand_path
from evariste.errors import EvaristeError

class CacheError(EvaristeError):
    """Exception related to cache."""
    pass

class AbstractCache:
    """Parent class of all cache classes."""

    #: Version of cache format. If the version read from cache is different
    #: from this, the cache is discarded.
    version = 1

    def __init__(self, *, setup, builder, **data):
        self.shared = Shared(setup=setup, builder=builder, **data)

    def close(self):
        """Close cache."""
        raise NotImplementedError

    @classmethod
    def clear(cls, cachedir):
        """Delete cache.

        :arg str cachedir: Cache directory.

        This is a class method, so that it can be called without the class
        being instanciated; i.e. cache can be deleted even if it is corrupted
        and cannot be loaded.
        """
        raise NotImplementedError


class NoCache(AbstractCache):
    """Fake cache: can be used if no cache is used.
    """

    def close(self):
        pass

    @classmethod
    def clear(cls, cachedir=None):
        pass

class Cache(AbstractCache):
    """Read cache."""

    dataname = ['tree', 'plugin']

    def __init__(self, cachedir, *, setup=None, builder=None):
        self.cachedir = expand_path(cachedir)
        os.makedirs(self.cachedir, exist_ok=True)

        data = {}

        if self._read_cache_version() == self.version:
            for base in self.dataname:
                data[base] = self._read_data(base)

        super().__init__(setup=setup, builder=builder, **data)

    def _read_cache_version(self):
        """Return cache version. Return 0 if an error occured."""
        try:
            with open(self._data_filename("version"), 'r') as file:
                return int(file.read())
        except (ValueError, FileNotFoundError):
            return 0

    def close(self):
        """Close cache: write data"""
        for base in self.dataname:
            self._write_data(
                base,
                getattr(self.shared, base),
                )
        with open(self._data_filename("version"), 'w') as file:
            file.write(str(self.version))

    def _data_filename(self, base):
        """Return the filename corresponding to data ``base``."""
        return os.path.join(self.cachedir, '{}.data'.format(base))

    def _read_data(self, base):
        """Read data ``base``, and return its content.

        If data could not be read, return a default value.
        """
        try:
            with open(self._data_filename(base), 'rb') as file:
                return SharedDeepDict.from_attr(base, pickle.load(file))
        except (FileNotFoundError, ValueError, EOFError, pickle.PickleError):
            # File does not exist, or data is corrupted: provide default data.
            return SharedDeepDict.from_attr(base)

    def _write_data(self, base, data):
        """Write data `base`."""
        with open(self._data_filename(base), 'wb') as file:
            pickle.dump(vars(data), file)

    @classmethod
    def clear(cls, cachedir):
        data = [
            os.path.join(cachedir, "{}.data".format(base))
            for base
            in ['version', 'plugin', 'tree']
            ]
        try:
            for path in data:
                try:
                    os.remove(path)
                except FileNotFoundError:
                    pass
            os.rmdir(cachedir)
        except OSError as error:
            raise CacheError("Unable to delete '{}': {}.".format(
                error.filename,
                str(error),
                ))

def open_cache(cachedir, setup, builder):
    """Return a cache object, possibly fake if `cachedir` is None."""
    if cachedir is None:
        return NoCache(setup=setup, builder=builder)
    else:
        return Cache(cachedir, setup=setup, builder=builder)



#!/usr/bin python

# Copyright 2015 Louis Paternault
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests"""

import doctest

from evariste import utils
import evariste

def load_tests(__loader, tests, __pattern):
    """Load tests (doctests).
    """
    # Loading doctests
    tests.addTests(doctest.DocTestSuite(evariste))
    for module in utils.iter_modules(
            evariste.__path__,
            prefix="{}.".format(evariste.__name__),
        ):
        try:
            tests.addTests(doctest.DocTestSuite(module))
        except ValueError:
            # No docstring, or no doctests in the docstrings
            pass

    return tests



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Render tree as an HTML (body) page, with CSS"""

import os
import pkg_resources

from evariste import utils
from evariste.renderer.html import HTMLRenderer

class HTMLBoxRenderer(HTMLRenderer):
    """Render tree as an HTML div (without the `<div>` tags)."""
    # pylint: disable=too-few-public-methods

    keyword = "htmlbox"
    depends = {
        'required': {
            'misc': [
                'copy',
            ],
        },
        'suggested': {
            'renderer.htmlbox.target': [
                'image',
                'default',
                ],
            'renderer.htmlbox.readme': [
                'html',
                ],
            },
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.environment.filters['labelize'] = self._labelize
        if 'copy_htmlbox' not in self.shared.setup['misc.copy']:
            self.shared.setup['misc.copy']['copy_htmlbox'] = (
                pkg_resources.resource_filename(
                    self.__class__.__module__,
                    os.path.join("data", "static"),
                    ),
                utils.expand_path(self.local.setup['staticdir']),
                )

    @staticmethod
    def _labelize(string):
        """Return a label, given a string.

        The label starts with alphabetical ascii characters, followed by
        digits. Two different strings have different hashes.
        """
        return "label" + str(hash(string))

    def _templatedirs(self):
        yield from super()._templatedirs()
        yield pkg_resources.resource_filename(
            "evariste.renderer.html",
            os.path.join("data", "templates"),
            )



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Access to git-versionned files."""

from datetime import datetime
import os
import pygit2

from evariste.vcs import VCS, NoRepositoryError

class Git(VCS):
    """Access git-versionned files"""
    # pylint: disable=no-member

    keyword = "git"

    def __init__(self, shared):
        super().__init__(shared)
        try:
            self.repository = pygit2.Repository(
                pygit2.discover_repository(
                    str(self.source)
                    )
                )
        except KeyError:
            raise NoRepositoryError(self.keyword, self.source)

        self._iter = None
        self._cached_mtime = None

    def walk(self):
        source = self.source.resolve().as_posix()
        for entry in self.repository.index:
            path = os.path.join(
                self.repository.workdir,
                entry.path,
                )
            if path.startswith(source):
                yield os.path.relpath(path, source)

    def is_versionned(self, path):
        return self.from_repo(path) in self.repository.index

    @property
    def workdir(self):
        return self.repository.workdir

    def _iter_start(self):
        """Initialize the commit iteration process."""
        self._iter = dict()
        self._iter["walker"] = self.repository.walk(
            self.repository.head.target,
            pygit2.GIT_SORT_TIME,
            )
        self._iter["commit"] = next(self._iter["walker"])

        self._cached_mtime = {
            entry.path: None
            for entry
            in self.repository.index
            }

    def _iter_walk(self):
        """Does one step of iterating over past commits, to set modification times"""
        old_commit = self._iter["commit"]
        self._iter["commit"] = next(self._iter["walker"])
        files_in_diff = (
            x.new_file_path
            for x in self._iter["commit"].tree.diff_to_tree(old_commit.tree)
            )
        for path in files_in_diff:
            if path in self._cached_mtime:
                if self._cached_mtime[path] is None:
                    self._cached_mtime[path] = datetime.fromtimestamp(old_commit.commit_time)

    def last_modified(self, path):
        if not self.is_versionned(path):
            return super().last_modified(path)

        if self.repository.status_file(self.from_repo(path)) == pygit2.GIT_STATUS_INDEX_NEW:
            # `path` has been added, but not committed
            return super().last_modified(path)

        path = self.from_repo(path)
        if self._cached_mtime is None:
            self._iter_start()

        # While modification time of `path` is not found, go back in time
        # (commit after commit) until we find a commit involving `path`.
        # Modification times of other files encountered on the way are stored,
        # so that they are retrieved faster the next time.
        while self._cached_mtime[path] is None:
            self._iter_walk()
        return self._cached_mtime[path]




# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract class for a renderer with a destination folder."""

import os

from evariste import errors, utils
from evariste.renderer import Renderer

class DestRenderer(Renderer):
    """Abstract class for a renderer with a destination folder."""
    # pylint: disable=too-few-public-methods

    def __init__(self, shared):
        super().__init__(shared)

        if self.local.setup['destdir'] is None:
            self.destdir = self.keyword
        else:
            self.destdir = utils.expand_path(self.local.setup['destdir'])
        try:
            os.makedirs(self.destdir, exist_ok=True)
        except FileExistsError:
            raise errors.EvaristeError("Cannot create directory '{}'.".format(self.destdir))

    def render(self, tree):
        raise NotImplementedError()



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Errors and exceptions"""

class EvaristeError(Exception):
    """Generic error for evariste"""

    def __init__(self, msg=""):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return self.msg

class EvaristeBug(Exception):
    """Bug: end-users should not see this."""
    def __init__(self, msg=""):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return self.msg



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Render images as HTML"""

from evariste.renderer.html.target import HtmlRenderer

IMAGE_EXTENSIONS = [
    "png",
    "jpg",
    "jpeg",
    "gif",
    ]

class HtmlImageRenderer(HtmlRenderer):
    """Render images as HTML"""

    priority = 0
    keyword = "image"

    def match(self, path):
        if not path.report.success:
            return False
        return path.report.target.as_posix().split('.')[-1].lower() in IMAGE_EXTENSIONS

    def render(self, tree, context):
        """Render ``tree``, which is a :class:`tree.File`, as HTML."""
        return context['render'](
            context,
            tree,
            ["tree_file_image.{}".format(self.extension)],
            )



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Raw action: does nothing; simply provides a link to download file."""

from evariste.action import Action, PathCompiler, Report

class RawCompiler(PathCompiler):
    """Raw compiler: does nothing to the file."""
    # pylint: disable=too-few-public-methods

    def compile(self):
        return Report(
            self.path,
            target=self.path.from_source,
            success=True,
            )


class Raw(Action):
    """Raw action"""

    keyword = "raw"
    priority = -float("inf")
    pathcompiler = RawCompiler

    def match(self, value):
        return True



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the plugin dependencies"""

import unittest

from evariste.builder import Builder
from evariste.setup import Setup

class TestLoadedPlugins(unittest.TestCase):
    """Load plugins"""

    default_plugins = set([
        'action.directory',
        'action.useless',
        'misc.changed',
        'vcs.none',
        ])

    @staticmethod
    def _loaded_plugins(local):
        """Create a builder, from a dummy setup file.

        Return a set a loaded plugins.
        """
        setup = Setup.from_dict({
            'setup': {
                'vcs': 'none',
                'source': ".",
                },
            'vcs': {
                'enable_plugins': ['none'],
                },
            'misc': {
                'disable_plugins': ['goodstuff'],
                },
            })
        setup.update(local)

        with Builder.from_setupdict(setup) as builder:
            return set(builder.plugins.iter_pluginid())



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Access to VCS (git, etc.) versionned files."""

from datetime import datetime
import os
import pathlib

from evariste import errors, plugins
from evariste.tree import Root

class NoRepositoryError(errors.EvaristeError):
    """No repository contains the given path."""

    def __init__(self, vcstype, directory):
        super().__init__()
        self.directory = directory
        self.vcstype = vcstype

    def __str__(self):
        return (
            "Could not find any {} repository containing directory '{}'."
            ).format(
                self.vcstype,
                self.directory,
                )

class VCS(plugins.PluginBase):
    """Generic class to access to versionned files."""

    plugin_type = "vcs"
    global_default_setup = {'setup': {'source': '.'}}

    def __init__(self, shared):
        super().__init__(shared)

    @property
    def source(self):
        """Return an absolute version of source setup option."""
        return pathlib.Path(self.shared.setup['setup']['source'])

    def walk(self):
        """Iterate over list of versionned files, descendants of source (as defined by setup file).
        """
        raise NotImplementedError()

    def is_versionned(self, path):
        """Return ``True`` iff ``path`` is versionned."""
        raise NotImplementedError()

    @property
    def workdir(self):
        """Return path of the root of the repository."""
        raise NotImplementedError()

    def from_repo(self, path):
        """Return ``path``, relative to the repository root."""
        return os.path.relpath(
            path.resolve().as_posix(),
            self.workdir
            )


    def last_modified(self, path):
        """Return the datetime of last modification."""
        # pylint: disable=no-self-use
        return datetime.fromtimestamp(os.path.getmtime(str(path)))



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""`evs cache` command: cache manipulation."""

import argparse
import logging
import os
import sys
import textwrap

from evariste.cache import Cache
from evariste.errors import EvaristeError
from evariste.setup import Setup
from evariste.utils import expand_path

LOGGER = logging.getLogger(__name__)

def _cache_type(name):
    """Check the argument and return its value.

    The argument must be an existing file or directory.
    """
    if not os.path.exists(name):
        raise argparse.ArgumentTypeError(
            "File '{}' does not exist.".format(name)
            )
    return name

def commandline_parser():
    """Return a command line parser."""

    parser = argparse.ArgumentParser(
        prog="evs cache",
        description="Perform operations on cache.",
        formatter_class=argparse.RawTextHelpFormatter,
        )

    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument(
        'cache',
        metavar="CACHE",
        help=textwrap.dedent("""
            Cache to process.

            If neither '--setup' nor '--dir' options are provided, implicit '--setup' is used if argument is a file, and implicit '--dir' is used if argument is a directory.
            """),
        type=_cache_type,
        )
    group = parent.add_mutually_exclusive_group()
    group.add_argument(
        '-s',
        '--setup',
        help="Argument is considered to be a setup file; cache directory is deduced from it.",
        action='store_true',
        )
    group.add_argument(
        '-d',
        '--dir',
        help="Argument is considered to be the cache directory.",
        action='store_true',
        )

    subparsers = parser.add_subparsers(
        description="",
        dest="command",
        )
    subparsers.required = True

    subparsers.add_parser(
        "clear",
        description="Delete cache.",
        parents=[parent],
        help="Delete cache.",
        )

    subparsers.add_parser(
        "console",
        description="Load cache and start an interactive console, with access to this cache.",
        parents=[parent],
        help="Load cache and start an interactive console.",
        )

    subparsers.add_parser(
        "dump",
        description="Dump cache. For more control over cache, see the 'console' subcommand.",
        parents=[parent],
        help="Dump cache (print its reprensentation)."
        )

    return parser

def cache_path(options):
    """Return the path of the cache."""
    if options.setup:
        setup = Setup.from_file(options.cache)
        if setup['setup']['cachedir'] is None:
            return os.path.join(
                os.path.dirname(options.cache),
                '.{}.cache'.format(os.path.basename(options.cache)),
                )
        else:
            return os.path.join(
                os.path.dirname(options.cache),
                expand_path(setup['setup']['cachedir']),
                )
    else:
        return options.cache

def get_cache_class(options):
    """Return the :class:`Cache` class corresponding to the setup file."""
    # pylint: disable=unused-argument
    return Cache

def get_cache(options):
    """Return the :class:`Cache` object corresponding to the setup file."""
    if options.setup:
        setup = Setup.from_file(options.cache)
    else:
        setup = None

    return get_cache_class(options)(
        cachedir=cache_path(options),
        setup=setup,
        )

def do_console(namespace):
    """Execute the `evs cache console` command."""
    cache = get_cache(namespace)

    try:
        import ipdb as pdb
    except ImportError:
        print("""You can install the `ipdb` package to get an enhanced Python debugger.""")
        import pdb

    # pylint: disable=unused-variable
    version = cache.version
    plugin = cache.shared.plugin
    tree = cache.shared.tree

    print(textwrap.dedent("""\
        You have access to the following variables. You can modify value of variables marked as 'read-only', but their value will not be written into cache:
        - cache: the `Cache` object.
        - version: the cache version (read-only).
        - plugin: the cached plugin information.
        - tree: the cached tree information.

        By default, nothing is written into cache. Use `cache.close()` if you want to. It is possible to write malformed cache data, which cannot be read again. If so, you can use `evs cache clear` to remove cache completely.

        Type of variables `plugin` and `tree` are custom types. To get a read-only version as a dictonary, use `vars(plugin)` and `vars(tree)`.

        Use `Ctrl-D` or `exit()` to leave the debugger.
        """))

    pdb.set_trace()

def do_dump(namespace):
    """Execute the `evs cache dump` command."""
    cache = get_cache(namespace)

    dump = {'version': cache.version}
    dump.update({
        base: getattr(cache.shared, base)
        for base
        in cache.dataname
        })

    from pprint import pprint
    pprint(dump)

def do_clear(namespace):
    """Execute the `evs cache clear` command."""
    get_cache_class(namespace).clear(cache_path(namespace))

COMMANDS = {
    'console': do_console,
    'dump': do_dump,
    'clear': do_clear,
    }

def parse_options():
    """Parse options, and perform some post-processing."""
    argumentparser = commandline_parser()
    options = argumentparser.parse_args()

    # Processing --setup/--dir options
    if options.setup:
        if not os.path.isfile(options.cache):
            argumentparser.error((
                "Argument '{}' provided with options '--setup' must be a file."
                ).format(options.cache))
    if options.dir:
        if not os.path.isdir(options.cache):
            argumentparser.error((
                "Argument '{}' provided with options '--dir' must be a directory."
                ).format(options.cache))
    if not (options.setup or options.dir):
        options.setup = os.path.isfile(options.cache)
        options.dir = os.path.isdir(options.cache)

    return options

def main():
    """Main option: run from command line."""
    options = parse_options()
    try:
        COMMANDS[options.command](options)
    except EvaristeError as error:
        LOGGER.error(str(error))
        sys.exit(1)



#!/bin/env python3

# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Command line client to :mod:`evs`"""

import argparse
import os
import textwrap
import glob
import sys
import re

import evariste

COMMAND_RE = re.compile(r'.*evs-(?P<command>\w*)')

def get_commands():
    """Return a dictionary of available commands, mapped to their corresponding binary.
    """
    commands = {}

    pathlist = os.environ['PATH'].split(":")
    for path in pathlist:
        path = os.path.expanduser(os.path.expandvars(path.strip()))
        for program in glob.iglob(os.path.join(path, "evs-*")):
            if os.path.isfile(program) and os.access(program, os.X_OK):
                match = COMMAND_RE.match(program)
                if match:
                    commands[match.groupdict()['command']] = program
    return commands

def print_error(error):
    """Print an error message."""
    print(error, "Try '--help' for more information.")

def print_version():
    """Print program version."""
    print(evariste.VERSION)

def print_help():
    """Print help text."""
    print(textwrap.dedent("""\
        usage: evs [-h] [--version] COMMAND [ARGUMENTS]

        Helper script for the evariste tool.

        positional arguments:
            COMMAND             Command to run. Available commands are: {commands}.

        optional arguments:
          -h, --help            show this help message and exit
          --version             Show version
        """).format(
            commands=", ".join(sorted(set(get_commands().keys()))),
        ))


def main():
    """Main function"""
    arguments = sys.argv[1:]

    if len(arguments) == 0:
        print_error("Not enough arguments.")
        sys.exit(1)

    if arguments[0] in ['-h', '--help']:
        print_help()
        sys.exit(0)

    if arguments[0] == '--version':
        print_version()
        sys.exit(0)

    commands = get_commands()
    if arguments[0] in commands:
        os.execvp(
            commands[arguments[0]],
            arguments,
            )

    print_error("Unrecognized command '{}'.".format(arguments[0]))
    sys.exit(1)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the plugin dependencies"""

import os

from evariste.plugins import PluginNotFound

from . import TestLoadedPlugins

DEPENDSDIR = os.path.join(
    os.path.dirname(__file__),
    "depends",
    )

def string2set(string):
    """Turn a string into a set of stripped non-empty lines"""
    return set([
        line.strip() for line in string.split("\n") if line.strip()
        ])

class TestDefaultSetup(TestLoadedPlugins):
    """Testing default setup"""
    # pylint: disable=too-few-public-methods

    def test_noplugins(self):
        """Test default and required plugins."""
        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    }),
                set([
                    'misc.default',
                    'misc.required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff',
                        },
                    }),
                set([
                    'misc.required',
                    ]) | self.default_plugins
                )

    def test_disable_required(self):
        """Exception is raised when disabling required plugin."""
        with self.assertLogs("evariste", level="WARNING") as log:
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'required goodstuff default',
                        },
                    }),
                set([
                    "misc.required",
                    ]) | self.default_plugins,
                )
            self.assertEqual(
                log.output,
                [(
                    "WARNING:evariste:Disabling plugin 'misc.required' is "
                    "asked by setup, but it is required. Still enabling."
                    )],
                )

    def test_non_existent_plugin(self):
        """Test exception when trying to enable a non-existent plugin."""
        with self.assertRaises(PluginNotFound):
            self._loaded_plugins({
                'setup': {
                    'libdirs': DEPENDSDIR,
                    },
                'misc': {
                    'enable_plugins': 'nonexistent',
                    },
                })

    def test_depends_simple(self):
        """Test simple dependencies"""
        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff',
                        'enable_plugins': 'dependsoptional',
                        },
                    }),
                set([
                    'misc.required',
                    'misc.dependsoptional',
                    'misc.foo',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff foo',
                        'enable_plugins': 'dependsoptional',
                        },
                    }),
                set([
                    'misc.required',
                    'misc.dependsoptional',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff',
                        'enable_plugins': 'dependsrequired',
                        },
                    }),
                set([
                    'misc.dependsrequired',
                    'misc.foo',
                    'misc.required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff foo',
                        'enable_plugins': 'dependsrequired',
                        },
                    }),
                set([
                    'misc.dependsrequired',
                    'misc.foo',
                    'misc.required',
                    ]) | self.default_plugins
                )

    def test_depends_recursive(self):
        """Test recursive dependencies"""
        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff foo',
                        'enable_plugins': 'dependsrequiredrequired',
                        },
                    }),
                set([
                    'misc.dependsrequiredrequired',
                    'misc.dependsrequired',
                    'misc.foo',
                    'misc.required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff',
                        'enable_plugins': 'dependsrequiredoptional',
                        },
                    }),
                set([
                    'misc.dependsrequiredoptional',
                    'misc.dependsoptional',
                    'misc.foo',
                    'misc.required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff foo',
                        'enable_plugins': 'dependsrequiredoptional',
                        },
                    }),
                set([
                    'misc.dependsrequiredoptional',
                    'misc.dependsoptional',
                    'misc.required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff',
                        'enable_plugins': 'dependsoptionalrequired',
                        },
                    }),
                set([
                    'misc.dependsoptionalrequired',
                    'misc.dependsrequired',
                    'misc.foo',
                    'misc.required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff dependsrequired',
                        'enable_plugins': 'dependsoptionalrequired',
                        },
                    }),
                set([
                    'misc.dependsoptionalrequired',
                    'misc.required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff',
                        'enable_plugins': 'dependsoptionaloptional',
                        },
                    }),
                set([
                    'misc.dependsoptional',
                    'misc.dependsoptionaloptional',
                    'misc.foo',
                    'misc.required',
                    ]) | self.default_plugins
                )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins({
                    'setup': {
                        'libdirs': DEPENDSDIR,
                        },
                    'misc': {
                        'disable_plugins': 'default goodstuff foo',
                        'enable_plugins': 'dependsoptionaloptional',
                        },
                    }),
                set([
                    'misc.dependsoptional',
                    'misc.dependsoptionaloptional',
                    'misc.required',
                    ]) | self.default_plugins
                )



#!/usr/bin/env python3

# Copyright 2015 Louis Paternault
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Installer"""

from setuptools import setup, find_packages
import codecs
import glob
import os
import sys

def readme():
    directory = os.path.dirname(os.path.join(
        os.getcwd(),
        __file__,
        ))
    with codecs.open(
        os.path.join(directory, "README.rst"),
        encoding="utf8",
        mode="r",
        errors="replace",
        ) as file:
        return file.read()

def evs_scripts():
    evs_path = os.path.join(os.path.dirname(sys.argv[0]), "evariste", "evs")
    for name in glob.iglob(os.path.join(evs_path, "*")):
        if os.path.isdir(name):
            name = os.path.basename(name)
        elif name.endswith(".py"):
            name = os.path.basename(name)[0:-3]
        else:
            continue
        bin = os.path.join(
            "bin",
            "evs-{}".format(name),
            )
        if os.path.isfile(bin) and os.access(bin, os.X_OK):
            yield os.path.basename(bin), name

setup(
        name='Evariste',
        version="0.1.0",
        packages=find_packages(exclude=["test*"]),
        setup_requires=["hgtools"],
        install_requires=[
            "jinja2",
            "docutils",
            "pygit2==0.22",
            "unidecode",
            ],
        include_package_data=True,
        author='Louis Paternault',
        author_email='spalax@gresille.org',
        description='TODO',
        url='https://git.framasoft.org/spalax/evariste',
        license="AGPLv3 or any later version",
        test_suite="test.suite",
        entry_points={
            'console_scripts':
                [
                    'evariste = evariste.main:main',
                    'evs = evariste.evs:main',
                ]
                +
                [
                    '{} = evariste.evs.{}:main'.format(bin, name)
                    for bin, name
                    in evs_scripts()
                    ]
                },
        classifiers=[ # TODO
            "Development Status :: 1 - Planning",
            "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        ],
        long_description=readme(),
        zip_safe = False,
)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Command line options"""

import argparse
import logging
import os
import textwrap

import evariste
from evariste import VERSION
from evariste.setup import Setup

LOGGER = logging.getLogger(evariste.__name__)
LOGGING_LEVELS = {
    -1: 100, # Quiet
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
    }

def _get_logging_level(verbose):
    """Turn command line verbosity into :mod:`LOGGING` verbosity."""
    if verbose in LOGGING_LEVELS:
        return LOGGING_LEVELS[verbose]
    if verbose is None:
        return LOGGING_LEVELS[0]
    if verbose > max(LOGGING_LEVELS.keys()):
        return LOGGING_LEVELS[max(LOGGING_LEVELS.keys())]
    raise NotImplementedError()

def _setup_type(name):
    """Check the argument and return its value.

    The argument must be an existing file.
    """
    if not os.path.exists(name):
        raise argparse.ArgumentTypeError(
            "File '{}' does not exist.".format(name)
            )
    if not os.path.isfile(name):
        raise argparse.ArgumentTypeError(
            "File '{}' is not a file.".format(name)
            )
    return name

def _itemgetint(value):
    """Return the first value of list, converted to an int."""
    return int(value[0])

class Options(argparse.Namespace):
    """Namespace of command line options.

    Added ability to iterate over option names.
    """
    # pylint: disable=too-few-public-methods

    def __str__(self):
        return str(dict([
            (attr, getattr(self, attr))
            for attr in dir(self)
            if not attr.startswith("_")
            ]))

    def __iter__(self):
        for attr in dir(self):
            if not attr.startswith("_"):
                yield attr

class Default:
    """Choose appropriate value given command line, setup file, etc.

    - If user set a value in the command line, it is used.
    - Othewise, if value is set in the ``arguments`` section of the setup file,
      use it.
    - At last, use the hard-coded default value defined in this module.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, name, defaultvalue, process_setup, process_arg):
        self.name = name
        self.defaultvalue = defaultvalue
        self.process_setup = process_setup
        self.process_arg = process_arg

    def __repr__(self):
        return "{}({}, {}, {}, {})".format(
            self.__class__.__name__,
            self.name,
            self.defaultvalue,
            self.process_setup,
            self.process_arg,
            )

    def get_value(self, value, setup):
        """Return the correct value, given the default values.

        :param str value: Value provided by the user.
        :param dict setup: The ``arguments`` section of the setup file, as a
            dictionary.
        """
        if value is not None:
            return self.process_arg(value)
        if self.name in setup:
            return self.process_setup(setup[self.name])
        return self.defaultvalue

class ArgumentParser(argparse.ArgumentParser):
    """Argument parser, using values from setup file as default values."""

    def __init__(self, *args, **kwargs):
        self.default = {}
        super().__init__(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        """This function adds two arguments to the value of its parent class.

        :param function process_arg: Function used to process command line
            argument (e.g. if argument is to be interpreted as an integer, this
            can be :func:`int`).
        :param function process_setup: Same usage, but is given as argument the
            value read from the setup file.
        """
        default = kwargs.pop("default", None)
        process_arg = kwargs.pop("process_arg", lambda x: x)
        process_setup = kwargs.pop("process_setup", lambda x: x)
        kwargs['default'] = None

        action = super().add_argument(*args, **kwargs)
        self.default[action.dest] = Default(action.dest, default, process_setup, process_arg)
        return action

    def parse_args(self, *args, **kwargs):
        """Parse user arguments, using setup options as default values."""
        namespace = super().parse_args(*args, **kwargs)

        # Process setup
        setup = vars(Setup.from_file(namespace.setup)).get('arguments', dict())

        # Set default values
        for attr in namespace:
            setattr(
                namespace,
                attr,
                self.default[attr].get_value(
                    getattr(namespace, attr),
                    setup=setup,
                    ),
                )

        # Process --verbose and --quiet
        if namespace.quiet:
            namespace.verbose = -1
        LOGGER.setLevel(_get_logging_level(namespace.verbose))

        return namespace

def commandline_parser():
    """Return a command line parser."""

    parser = ArgumentParser(
        description=(
            "Recursively compile files in a directory, and render result."
            ),
        formatter_class=argparse.RawTextHelpFormatter,
        )

    parser.add_argument(
        '--version',
        help='Show version',
        action='version',
        version='%(prog)s ' + VERSION
        )

    parser.add_argument(
        "-v", "--verbose",
        help='Verbose. Repeat for more details.',
        action='count',
        )

    parser.add_argument(
        "-q", "--quiet",
        help='Quiet. Does not print anything to standard output.',
        action='store_true',
        default=False,
        )

    parser.add_argument(
        "-j", "--jobs",
        help=(
            "Specify the number of jobs to run simultaneously. Default is one "
            "more than the number of CPUs. "
            ),
        action='store',
        nargs=1,
        default=os.cpu_count()+1,
        process_setup=int,
        process_arg=_itemgetint,
        )

    parser.add_argument(
        "-B", "--always-compile",
        dest='always_compile',
        help='Unconditionally make all targets',
        action='store_true',
        default=False,
        process_setup=bool,
        )

    parser.add_argument(
        'setup',
        metavar="SETUP",
        help=textwrap.dedent("""
            Setup file to process.
            """),
        type=_setup_type,
        )

    return parser

def get_options():
    """Return the namespace of command line options."""
    return commandline_parser().parse_args(namespace=Options())



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the shared plugin"""

import unittest

from evariste.builder import Builder

def default_shared():
    """Return a default, as empty as possible, :class:`evariste.shared.Shared` object.
    """
    return Builder.from_setupdict(
        {
            'setup': {
                'vcs': 'none',
                },
            'vcs': {
                'enable_plugins': ['none'],
                },
            }
        ).shared

class TestShared(unittest.TestCase):
    """Testing Shared"""
    # pylint: disable=too-few-public-methods, no-member

    def test_setup(self):
        """Test Shared.setup"""
        shared = default_shared()

        self.assertEqual(shared.setup['plugin']['foo'], None)
        shared.setup['plugin']['foo'] = 'bar'
        self.assertEqual(shared.setup['plugin']['foo'], 'bar')

        view = shared.get_plugin_view('plugin')
        self.assertEqual(view.setup['foo'], 'bar')
        self.assertEqual(view.setup['foo1'], None)
        view.setup['foo1'] = 'bar1'
        self.assertEqual(view.setup['foo1'], 'bar1')
        self.assertEqual(shared.setup['plugin']['foo1'], 'bar1')

    def test_plugin(self):
        """Test Shared.plugin"""
        shared = default_shared()

        self.assertEqual(shared.plugin['plugin1'], None)
        shared.plugin['plugin1'] = 'bar1'
        self.assertEqual(shared.plugin['plugin1'], 'bar1')

        view1 = shared.get_plugin_view('plugin1')
        view2 = shared.get_plugin_view('plugin2')
        self.assertEqual(view1.plugin, 'bar1')
        self.assertEqual(view2.plugin, None)
        view2.plugin = 'bar2'
        self.assertEqual(view2.plugin, 'bar2')
        self.assertEqual(shared.plugin['plugin2'], 'bar2')

    def test_tree(self):
        """Test Shared.tree"""
        shared = default_shared()

        self.assertEqual(shared.tree['path1']['plugin1'], None)
        shared.tree['path1']['plugin1'] = 'bar1'
        self.assertEqual(shared.tree['path1']['plugin1'], 'bar1')

        view = shared.get_tree_view('path1')
        self.assertEqual(view.tree['plugin1'], 'bar1')
        self.assertEqual(view.tree['plugin2'], None)
        view.tree['plugin2'] = 'bar2'
        self.assertEqual(view.tree['plugin2'], 'bar2')
        self.assertEqual(shared.tree['path1']['plugin2'], 'bar2')

        plugin_view = shared.get_plugin_view('plugin1')
        self.assertEqual(plugin_view.tree['path1'], 'bar1')

        tree2_view = shared.get_tree_view('path2')
        self.assertEqual(tree2_view.tree['plugin1'], None)
        self.assertEqual(plugin_view.tree['path2'], None)
        tree2_view.tree['plugin1'] = "bar"
        self.assertEqual(plugin_view.tree['path2'], "bar")
        self.assertEqual(tree2_view.tree['plugin1'], "bar")
        self.assertEqual(shared.tree['path2']['plugin1'], "bar")



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Builder

Takes care of build process.
"""

import functools
import logging
import os

from evariste import plugins
from evariste import multithread
from evariste.cache import open_cache
from evariste.hooks import methodhooks
from evariste.plugins import AllPlugins
from evariste.setup import Setup, SetupError
from evariste.tree import Root
import evariste

LOGGER = logging.getLogger(evariste.__name__)

class Builder:
    """Takes care of build process. Can be used as a context.
    """
    # pylint: disable=no-member

    def __init__(self, setup, cachedir=None):
        self.cache = open_cache(cachedir, setup, self)
        self.shared = self.cache.shared
        self.plugins = AllPlugins(shared=self.shared)
        self.lock = multithread.Lock()

        LOGGER.info("Building directory tree…")
        try:
            with self.plugins.contexthook("buildtree", args=[self]):
                self.tree = Root.from_vcs(
                    self.plugins.match_plugin('vcs', self.shared.setup['setup']['vcs']),
                    )
        except plugins.NoMatch as error:
            raise SetupError(
                "Setup error: Value '{}' is not a valid vcs (available ones are: {}).".format(
                    error.value,
                    ", ".join(["'{}'".format(item) for item in error.available]),
                    )
                )

    @methodhooks(name="compiletree")
    def compile(self):
        """Compile files handled by this builder."""
        LOGGER.info("Compiling…")
        self.tree.root_compile()
        self.prune_readmes()

    def prune_readmes(self):
        """Remove readme files from tree."""
        for plugin in self.plugins.get_plugin_type('renderer').values():
            plugin.prune_readmes(self.tree)

    @methodhooks()
    def close(self):
        """Perform close operations."""
        self.cache.close()

    @classmethod
    def from_setupname(cls, name, *, cachedir=None, default=None):
        """Factory that returns a builder, given the name of a setup file."""
        LOGGER.info("Reading setup…")
        setup = Setup.from_file(name).update(default)
        if cachedir is None:
            if setup['setup']['cachedir'] is not None:
                cachedir = setup['setup']['cachedir']
            else:
                cachedir = '.{}.cache'.format(os.path.basename(name))
        if default is None:
            default = {}
        return cls(
            setup,
            os.path.join(
                os.path.dirname(name),
                cachedir,
                ),
            )

    @classmethod
    def from_setupdict(cls, dictionary, *, cachedir=None, default=None):
        """Factory that returns a builder, given a setup dictionary."""
        LOGGER.info("Reading setup…")
        if default is None:
            default = {}
        if cachedir is None:
            try:
                cachedir = dictionary['setup']['cachedir']
            except KeyError:
                pass
        return cls(Setup.from_dict(dictionary).fill_blanks(default), cachedir)

    def renderer(self, keyword):
        """Returns a :func:`functools.partial` object over the `keyword` renderer.
        """
        return functools.partial(self.plugins.match_plugin("renderer", keyword).render, self.tree)

    def iter_renderers(self):
        """Iterator over renderers.

        Returns a :func:`functools.partial` object, over the renderer
        :meth:`render` method, with tree as the first argument. More arguments
        can be added when calling the partial object.
        """
        for keyword in self.shared.setup["renderer"]["enable_plugins"]: # pylint: disable=no-member
            LOGGER.info("Rendering {}…".format(keyword))
            yield self.renderer(keyword)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.close()
        return False
